/* IP.PRG */

/* This replicates some of the empirical work reported in 
"Testing for Linearity" by Bruce E. Hansen.
For updates and contact information, see my webpage
www.ssc.wisc.edu/~bhansen
*/

/*  This program estimates a two-regime TAR model, and tests the null of a 
one-regime AR against the alternative of a two-regime TAR */

yname="Industrial Production";	@ name of series @
output file = "ip.out" reset; @ name of output file @
load dat[]=ip.dat;
dat=ln(dat);
dat=(dat[13:rows(dat)]-dat[1:rows(dat)-12])*100;
dat=dat[157:rows(dat)];
		@ load and define univariate series @
p=16;		@ order of autoregression @
_dmin=1;		@ minimal delay order @
_dmax=p;		@ maximal delay order (set to p or less @
_trim = .1;		@ minimal percentage of data per regime @
_qnum=100;	@ Number of thresholds to search. Set _qnum=0 to search over all values @
_boot = 2000;	@ bootstrap replications @

/*********************************************************************/

@ Define Data @
n=rows(dat);
y=dat[p+1:n];
t=n-p;
x=ones(t,1)~dat[p:n-1];
xname="Constant"|"Y(t-1)";
xname2="Constant"|"Y(t-1)^2";
j=2; do while j<=p;
    x=x~dat[p+1-j:n-j];
    pn = ftocv(j,1,0)*(j<10)+ftocv(j,2,0)*(j>=10);
    namej = "" $+ "Y(t-" $+ pn $+ ")";
    xname=xname|namej;
    xname2=xname2|(namej $+ "^2");
j=j+1;endo;
k=cols(x);

@ Linear Regression @
mi=invpd(moment(x,0));
beta=mi*(x'y);
e=y-x*beta;
ee=e'e;
xe=x.*e;
sig=ee/t;
se=sqrt(diag(mi*moment(xe,0)*mi));

@ Conditional Variance @
e2=e.^2;
x2 = x.^2;
m2=invpd(moment(x2,0));
hetbeta=m2*(x2'e2);
h=x2*hetbeta;
eh=e./sqrt(h.*(h.>0)+(h.<=0)).*(h.>0);
u=e2-h;
se2=sqrt(diag(m2*moment(x2.*u,0)*m2));
em=e2-meanc(e2);
fh=t*((em'em)/(u'u)-1);

@ Report Linear Estimation @
ff = "#*.*lG"~20~8;
print ("Dependent Variable:     " $+ yname);
"";
"Linear Autoregression";
"";
pr = printfm("Variable"~"Estimate"~"St Error",0~0~0,ff);"";
"       __________________________________________";
pr = printfm(xname~beta~se,0~1~1,ff);"";
"";
"Observations                      " t;
"Sum of Squared Errors             " ee;
"Residual Variance                 " sig;
"";"";
"Conditional Variance";"";
pr = printfm("Variable"~"Estimate"~"St Error",0~0~0,ff);"";
"       __________________________________________";
pr = printfm(xname2~hetbeta~se2,0~1~1,ff);"";
"";
"Heteroskedasticity F Test  " fh~cdfchic(fh,p);"";
"       __________________________________________";
"";"";
"";"";


@ TAR Estimation @
q = qfunc(x);
qq=qsort(q);
qn=cols(q);gn=rows(qq);
s=zeros(gn,qn);
mmistore=zeros(k*(k+1)/2,qn*gn);
m=1; do while m<=qn;
 qm=q[.,m]; 
  j=1;do while j<=gn;
    gg=qq[j,m];
    if scalmiss(gg); continue; endif;
    xd=x.*(qm.<=gg);
    xxd=moment(xd,0);
    xde=xd'e;
    mmi=ginv(xxd-xxd*mi*xxd);
    mmistore[.,(m-1)*gn+j]=vech(mmi);
    s[j,m]=xde'mmi*xde;
  j=j+1;endo;
m=m+1;endo;
ghat=diag(qq[maxindc(s),.]);
f=maxc(s);
f=t*(f./(ee-f));
mf=maxc(f);
di=maxindc(f);
delay=_dmin-1+di;
gghat=ghat[di];

d1=(q[.,di].<=gghat);
d2=1-d1;
x1=x.*d1;
x2=x.*d2;
xx=x1~x2;
mxx=invpd(moment(xx,0));
betatar=mxx*(xx'y);
etar=y-xx*betatar;
xxe=xx.*etar;
eetar=etar'etar;
sigtar=eetar/t;
n1=sumc(d1);n2=sumc(d2);
setar=sqrt(diag(mxx*moment(xxe,0)*mxx));
sig1=sumc((etar.*d1).^2)/n1;
sig2=sumc((etar.*d2).^2)/n2;

e2=etar.^2;d12=d1~d2;
z=d12~(x[.,2:k].^2);
mz=invpd(moment(z,0));
betaz=mz*(z'e2);
u=e2-z*betaz;
se3=sqrt(diag(mz*moment(z.*u,0)*mz));
em=e2-meanc(e2);
fz=t*((em'em)/(u'u)-1);
xname3="D1"|"D2"|xname2[2:k];

@ Report TAR estimates @

"Threshold Autoregression";
"Sum of Squared Errors             " eetar;
"Residual Variance                 " sigtar;
"Delay Order                         " delay;
"Threshold Estimate               " gghat;
"";"";
"Regime 1 (Threshold Variable less than or equal to Threshold):";"";
pr = printfm("Variable"~"Estimate"~"St Error",0~0~0,ff);"";
"       __________________________________________";
pr = printfm(xname~betatar[1:k]~setar[1:k],0~1~1,ff);"";
"";
"Observations                      " n1;
"Percentage                         " (n1/t);
"Regime Variance                  " sig1;
"";"";
"Regime 2 (Threshold Variable greater than Threshold):";"";
pr = printfm("Variable"~"Estimate"~"St Error",0~0~0,ff);"";
"       __________________________________________";
pr = printfm(xname~betatar[k+1:2*k]~setar[k+1:2*k],0~1~1,ff);"";
"";
"Observations                      " n2;
"Percentage                         " (n2/t);
"Regime Variance                  " sig2;
"";"";"";
"Conditional Variance";"";
pr = printfm("Variable"~"Estimate"~"St Error",0~0~0,ff);"";
"       __________________________________________";
pr = printfm(xname3~betaz~se3,0~1~1,ff);"";
"";
"Heteroskedasticity F Test  " fz~cdfchic(fz,k);"";
"";
"       __________________________________________";
"";"";"";"";

@ P-Value Calculations @

@ Asymptotic Bootstrap @
  ab1=zeros(_boot,p);
  ab2=zeros(_boot,p);
i=1; do while i<=_boot;
  y1=rndn(t,1);
  e1=(y1-x*mi*(x'y1));
  s1=zeros(gn,qn);
  y2=e.*y1;
  e2=(y2-x*mi*(x'y2));
  s2=zeros(gn,qn);
  m=1; do while m<=qn;
   qm=q[.,m]; 
    j=1;do while j<=gn;
      gg=qq[j,m];
      if scalmiss(gg); continue; endif;
      xd=x.*(qm.<=gg);
      mmi=xpnd(mmistore[.,(m-1)*gn+j]);
      xde=xd'e1;
      s1[j,m]=xde'mmi*xde;
      xde=xd'e2;
      s2[j,m]=xde'mmi*xde;
    j=j+1;endo;
  m=m+1;endo;
  s1=maxc(s1);
  ab1[i,.]=s1';
  s2=maxc(s2);
  ab2[i,.]=(s2')./sig;
i=i+1;endo;
  ma1=maxc(ab1');
  apv1=meanc(ab1.>(f'));
  ampv1=meanc(ma1.>mf);
  ma2=maxc(ab2');
  apv2=meanc(ab2.>(f'));
  ampv2=meanc(ma2.>mf);
clear mmistore,xxdstore;


@ Model-Based Bootstrap @
y0=dat[1:p];
@ Homoskedastic Error Bootstrap @
  fb=zeros(_boot,p);
  i=1; do while i<=_boot;
    fb[i,.]=tar(ar_sim(y0,beta,e))';
  i=i+1;endo;
  mbt1=maxc(fb');
  btpv1=meanc(fb.>(f'));
  btmpv1=meanc(mbt1.>mf);
@ Heteroskedastic Error Bootstrap @
  fb=zeros(_boot,p);
  i=1; do while i<=_boot;
    fb[i,.]=tar(het_sim(y0,beta,hetbeta,eh))';
  i=i+1;endo;
  mbt2=maxc(fb');
  btpv2=meanc(fb.>(f'));
  btmpv2=meanc(mbt2.>mf);

@ Output @
"Bootstrap Replications      :   " _boot;
names="Delay"~"Fstat"~"PV-A-H"~"PV-B-H"~"PV-A-Het"~"PV-B-Het";
stat=seqa(1,1,p)~f~apv1~btpv1~apv2~btpv2;
statm="Max"~mf~ampv1~btmpv1~ampv2~btmpv2;
  "";"";
  ff = "#*.*lG"~10~8;
  "TAR Tests, P-Values";"";
  pr = printfm(names,zeros(1,cols(names)),ff);"";
  "       __________________________________________";
  ff = "#*.*lG"~10~4;
  pr = printfm(stat,ones(1,cols(stat)),ff);"";
  pr = printfm(statm,0~ones(1,cols(stat)-1),ff);"";
"";"";
output off;

@ Density Calculation @
qm=.99;
ub=maxc(quant(ma1,qm)|quant(ma2,qm)|quant(mbt1,qm)|quant(mbt2,qm));
k_x=seqa(0,ub/999,1000);
k1=(k_x.^((k/2)-1)).*exp(-k_x/2)./(gamma(k/2)*(2^(k/2)));
k1=k1~kernel(ma1,k_x)~kernel(mbt1,k_x)~kernel(ma2,k_x)~kernel(mbt2,k_x);
ip_k1=k_x~k1;
save ip_k1;

/*****************************************************/
@ TAR Procedures @
proc tar(dat);
local y,x,q,qq,mi,e,qn,gn,s,m,qm,j,xd,xde,f,xxd,gg,mmi;
  y=dat[p+1:n];
  x=ones(t,1)~dat[p:n-1];
  j=2; do while j<=p;
    x=x~dat[p+1-j:n-j];
  j=j+1;endo;
  q = qfunc(x);
  qq=qsort(q);
  mi=ginv(moment(x,0));
  e=y-x*mi*(x'y);
  qn=cols(q);gn=rows(qq);
  s=zeros(gn,qn);
  m=1; do while m<=qn;
   qm=q[.,m]; 
    j=1;do while j<=gn;
      gg=qq[j,m];
      if scalmiss(gg); continue; endif;
      xd=x.*(qm.<=gg);
      xxd=moment(xd,0);
      mmi=xxd-xxd*mi*xxd;
      xde=xd'e;
      s[j,m]=xde'dinv(xde,mmi);
    j=j+1;endo;
  m=m+1;endo;
  f=maxc(s);
  f=t*(f./((e'e)-f));
retp(f);
endp;

proc qsort(q);
local j,n,n1,qq,k,qj,nj;
  n=rows(q);k=cols(q);
  n1=round(_trim*n);
  if _qnum==0;
    qq=zeros(n-n1-n1+1,k);
    qq=miss(qq,0);
    j=1; do while j<=k;
      qj=unique(q[.,j],1);
      qj=qj[n1:(rows(qj)-n1+1)];
      qq[1:rows(qj),j]=qj;
    j=j+1;endo;
  else;
    qq=zeros(_qnum,k);
    qq=miss(qq,0);
    j=1; do while j<=k;
      qj=unique(q[.,j],1);
      qj=qj[n1:(rows(qj)-n1+1)];
      nj=rows(qj);
      if nj<=_qnum;
        qq[1:nj,j]=qj;
      else;
        qq[.,j]=qj[ceil(seqa(1,1,_qnum)./_qnum*nj)];
      endif;
    j=j+1;endo;
  endif;
retp(qq);
endp;

proc qfunc(x);
local q;
  q=x[.,(_dmin+1):(_dmax+1)];
retp(q);
endp;

/*****************************************************/
@ Simulation Procedures @

proc ar_sim(y0,beta,e);
local t,u,yb;
  t=rows(e);
  u=e[ceil(rndu(t,1)*t)]+beta[1];
  yb=recserar(y0|u,y0,beta[2:k]);
retp(yb);
endp;


proc het_sim(y0,beta,het,e);
local t,u,j,y,x,h;
  t=rows(e);
  u=e[ceil(rndu(t,1)*t)];
  y=y0|zeros(t,1);
  j=1; do while j<=t;
    x=1|rev(y[j:p+j-1]);
    h=(x.^2)'het;
    y[p+j]=beta'x+sqrt(h.*(h.>0)).*u[j];
  j=j+1;endo;
retp(y);
endp;

/*****************************************************/  

proc kernel(x,b);
local h,g,kern,i,u;
  h=1.7*stdc(x)/(rows(x)^(.2));
  g=rows(b);
  kern=zeros(g,1);
  i=1; do while i<=g;
    u=abs(b[i]-x)/h;
    kern[i]=meanc((1-u.^2).*(u.<=1))*(.75)/h;
  i=i+1;endo;
retp(kern);
endp;

proc quant(x,q);
local s,qq;
  s=sortc(x,1);
  qq=s[round(rows(s)*q)];
retp(qq);
endp;

proc ginv(m);
local mi;
  trap 1;
  mi=invpd(m);
  trap 0;
  if scalerr(mi);
    mi=pinv(m);
  endif;
retp(mi);
endp;

proc dinv(y,x);
local d;
  trap 1;
    d=y/x;
  trap 0;
  if scalerr(d);
    d=ginv(moment(x,0))*(x'y);
  endif;
retp(d);
endp;

/*****************************************************/  