/* IP3.PRG */

/* This replicates some of the empirical work reported in 
"Testing for Linearity" by Bruce E. Hansen.
For updates and contact information, see my webpage
www.ssc.wisc.edu/~bhansen
*/

/*  This program estimates a two-threshold TAR model, and tests the null of a 
 one-threshold TAR against the alternative of a two-threshold TAR */

yname="Industrial Production";	@ name of series @
output file = "ip3.out" reset; @ name of output file @
load dat[]=ip.dat;
dat=ln(dat);
dat=(dat[13:rows(dat)]-dat[1:rows(dat)-12])*100;
dat=dat[157:rows(dat)];
		@ load and define univariate series @
p=16;		@ order of autoregression @
_dmin=1;		@ minimal delay order @
_dmax=p;		@ maximal delay order (set to p or less @
_trim = .1;		@ minimal percentage of data per regime @
_qnum=100;	@ Number of thresholds to search. Set _qnum=0 to search over all values @
_boot = 2000;	@ bootstrap replications @

/*********************************************************************/

@ Define Data @
n=rows(dat);
y=dat[p+1:n];
t=n-p;
x=ones(t,1)~dat[p:n-1];
xname="Constant"|"Y(t-1)";
xname2="Constant"|"Y(t-1)^2";
j=2; do while j<=p;
    x=x~dat[p+1-j:n-j];
    pn = ftocv(j,1,0)*(j<10)+ftocv(j,2,0)*(j>=10);
    namej = "" $+ "Y(t-" $+ pn $+ ")";
    xname=xname|namej;
    xname2=xname2|(namej $+ "^2");
j=j+1;endo;
k=cols(x);


@ TAR Estimation @

{f,dhat,ghat,ghat1,ghat2} = tar2(dat);

q = qfunc(x);
qd=q[.,dhat];
delay=_dmin-1+dhat;

@ TAR_1 Model @

d1=(qd.<=ghat);
d2=1-d1;
xx=(x.*d1)~(x.*d2);
mxx=invpd(moment(xx,0));
beta=mxx*(xx'y);
e=y-xx*beta;
xxe=xx.*e;
ee=e'e;
sig=ee/t;
n1=sumc(d1);n2=sumc(d2);
se=sqrt(diag(mxx*moment(xxe,0)*mxx));
b1=beta[1:k];b2=beta[k+1:2*k];
e1=selif(e,d1);
e2=selif(e,d2);
sig1= (e1'e1)/n1;
sig2= (e2'e2)/n2;


yz=e.^2;d12=d1~d2;
z=d12~(x[.,2:k].^2);
mz=invpd(moment(z,0));
betaz=mz*(z'yz);
h=z*betaz;
u=yz-h;
se3=sqrt(diag(mz*moment(z.*u,0)*mz));
em=yz-meanc(yz);
fz=t*((em'em)/(u'u)-1);
xname3="D1"|"D2"|xname2[2:k];
ez=e./sqrt(h.*(h.>0)+(h.<=0)).*(h.>0);


ff = "#*.*lG"~20~8;
print ("Dependent Variable:     " $+ yname);"";
"Threshold Autoregression, Single Threshold";
"Sum of Squared Errors             " ee;
"Residual Variance                 " sig;
"Delay Order                         " delay;
"Threshold Estimate               " ghat;
"";"";
"Regime 1 (Threshold Variable less than or equal to Threshold):";"";
pr = printfm("Variable"~"Estimate"~"St Error",0~0~0,ff);"";
"       __________________________________________";
pr = printfm(xname~b1~se[1:k],0~1~1,ff);"";
"";
"Observations                      " n1;
"Percentage                         " (n1/t);
"Regime Variance                  " sig1;
"";"";
"Regime 2 (Threshold Variable greater than Threshold):";"";
pr = printfm("Variable"~"Estimate"~"St Error",0~0~0,ff);"";
"       __________________________________________";
pr = printfm(xname~b2~se[k+1:2*k],0~1~1,ff);"";
"";
"Observations                      " n2;
"Percentage                         " (n2/t);
"Regime Variance                  " sig2;
"";"";"";
"Conditional Variance";"";
pr = printfm("Variable"~"Estimate"~"St Error",0~0~0,ff);"";
"       __________________________________________";
pr = printfm(xname3~betaz~se3,0~1~1,ff);"";
"";
"Heteroskedasticity F Test  " fz~cdfchic(fz,k);"";
"";
"       __________________________________________";
"";"";"";"";


@ TAR_2 Model @

d1=(qd.<=ghat1);
d2=(qd.<=ghat2).*(1-d1);
d3=1-d1-d2;
xx=(x.*d1)~(x.*d2)~(x.*d3);
mxx=invpd(moment(xx,0));
betatar=mxx*(xx'y);
etar=y-xx*betatar;
xxe=xx.*etar;
eetar=etar'etar;
sigtar=eetar/t;
n1=sumc(d1);n2=sumc(d2);n3=sumc(d3);
setar=sqrt(diag(mxx*moment(xxe,0)*mxx));
sig1=sumc((etar.*d1).^2)/n1;
sig2=sumc((etar.*d2).^2)/n2;
sig3=sumc((etar.*d3).^2)/n3;


"Threshold Autoregression, Double Threshold";
"Sum of Squared Errors             " eetar;
"Residual Variance                 " sigtar;
"Delay Order                         " delay;
"First Threshold Estimate         " ghat1;
"Second Threshold Estimate       " ghat2;
"";"";
"Regime 1 (Threshold Variable less than or equal to First Threshold):";"";
pr = printfm("Variable"~"Estimate"~"St Error",0~0~0,ff);"";
"       __________________________________________";
pr = printfm(xname~betatar[1:k]~setar[1:k],0~1~1,ff);"";
"";
"Observations                      " n1;
"Percentage                         " (n1/t);
"Regime Variance                  " sig1;
"";"";
"Regime 2 (Threshold Variable Between Thresholds):";"";
pr = printfm("Variable"~"Estimate"~"St Error",0~0~0,ff);"";
"       __________________________________________";
pr = printfm(xname~betatar[k+1:2*k]~setar[k+1:2*k],0~1~1,ff);"";
"";
"Observations                      " n2;
"Percentage                         " (n2/t);
"Regime Variance                  " sig2;
"";"";
"Regime 3 (Threshold Variable Above Second Threshold):";"";
pr = printfm("Variable"~"Estimate"~"St Error",0~0~0,ff);"";
"       __________________________________________";
pr = printfm(xname~betatar[2*k+1:3*k]~setar[2*k+1:3*k],0~1~1,ff);"";
"";
"Observations                      " n3;
"Percentage                         " (n3/t);
"Regime Variance                  " sig3;
"       __________________________________________";
"";"";"";"";

 y0=dat[1:p];

@ Model-Based Bootstrap, Homoskedastic Errors @
  fb1=zeros(_boot,1);
  i=1; do while i<=_boot;
   {fbi,dhatb,ghatb,ghat1b,ghat2b} = tar2(tar_sim1(y0,b1,b2,delay,ghat,e));
    fb1[i]=fbi;
  i=i+1;endo;
  pv1=meanc(fb1.>f);

@ Model-Based Bootstrap, Regime-Specific Errors @
  fb2=zeros(_boot,1);
  i=1; do while i<=_boot;
   {fbi,dhatb,ghatb,ghat1b,ghat2b} = tar2(tar_sim2(y0,b1,b2,delay,ghat,e1,e2));
    fb2[i]=fbi;
  i=i+1;endo;
  pv2=meanc(fb2.>f);

@ Model-Based Bootstrap, Heteroskedastic Errors @
  fb3=zeros(_boot,1);
  i=1; do while i<=_boot;
   {fbi,dhatb,ghatb,ghat1b,ghat2b} = tar2(tar_sim2(y0,b1,b2,delay,ghat,betaz,ez));
    fb3[i]=fbi;
  i=i+1;endo;
  pv3=meanc(fb3.>f);

@ Output @
"Bootstrap Replications      :   " _boot;
names="Fstat"~"PV-1"~"PV-2"~"PV-3";
statm=f~pv1~pv2~pv3;
"";"";
ff = "#*.*lG"~12~7;
"TAR Tests, P-Values";"";
pr = printfm(names,0~0~0~0,ff);"";
"       __________________________________________";
ff = "#*.*lG"~12~4;
pr = printfm(statm,1~1~1~1,ff);"";
"";"";
output off;

@ Density Calculation @

ub=maxc(quant(fb1,.99)|quant(fb2,.99)|quant(fb3,.99));
k_x=seqa(0,ub/999,1000);
k_chi=(k_x.^((k/2)-1)).*exp(-k_x/2)./(gamma(k/2)*(2^(k/2)));
kern=k_chi~kernel(fb1,k_x)~kernel(fb2,k_x)~kernel(fb3,k_x);
ip_k3=k_x~kern;
save ip_k3;


/*****************************************************/
@ TAR Procedures @

proc (2) = tar(y,x,q,qq);
local mi,e,qn,gn,s,m,qm,j,xd,xde,xxd,gg,mmi,dhat,ghat;
  mi=ginv(moment(x,0));
  e=y-x*mi*(x'y);
  qn=cols(q);gn=rows(qq);
  s=zeros(gn,qn);
  m=1; do while m<=qn;
   qm=q[.,m]; 
    j=1;do while j<=gn;
      gg=qq[j,m];
      if scalmiss(gg); continue; endif;
      xd=x.*(qm.<=gg);
      xxd=moment(xd,0);
      mmi=xxd-xxd*mi*xxd;
      xde=xd'e;
      s[j,m]=xde'dinv(xde,mmi);
    j=j+1;endo;
  m=m+1;endo;
  dhat=maxindc(maxc(s));
  ghat=qq[maxindc(s[.,dhat]),dhat];
retp(dhat,ghat);
endp;

proc tar_d(y,x,x1,q,qq);
local mi,e,qn,gn,s,j,xd,xde,xxd,gg,mmi,ghat,xdx;
  mi=ginv(moment(x,0));
  e=y-x*mi*(x'y);
  gn=rows(qq);
  s=zeros(gn,1);
    j=1;do while j<=gn;
      gg=qq[j];
      if scalmiss(gg); continue; endif;
      xd=x1~(x.*(q.<=gg));
      xxd=moment(xd,0);
      xdx=x'xd;
      mmi=xxd-xdx'mi*xdx;
      xde=xd'e;
      s[j]=xde'dinv(xde,mmi);
    j=j+1;endo;
  ghat=qq[maxindc(s)];
retp(ghat);
endp;


proc qsort(q);
local j,n,n1,qq,k,qj,nj;
  n=rows(q);k=cols(q);
  n1=round(_trim*n);
  if _qnum==0;
    qq=zeros(n-n1-n1+1,k);
    qq=miss(qq,0);
    j=1; do while j<=k;
      qj=unique(q[.,j],1);
      qj=qj[n1:(rows(qj)-n1+1)];
      qq[1:rows(qj),j]=qj;
    j=j+1;endo;
  else;
    qq=zeros(_qnum,k);
    qq=miss(qq,0);
    j=1; do while j<=k;
      qj=unique(q[.,j],1);
      qj=qj[n1:(rows(qj)-n1+1)];
      nj=rows(qj);
      if nj<=_qnum;
        qq[1:nj,j]=qj;
      else;
        qq[.,j]=qj[ceil(seqa(1,1,_qnum)./_qnum*nj)];
      endif;
    j=j+1;endo;
  endif;
retp(qq);
endp;


proc qsort_2(q,ghat);
local n,n1,qq,nq;
  n=rows(q);
  n1=round(_trim*n);
  qq=unique(q,1);
  qq=qq[n1:(rows(qq)-n1+1)];
  qq=delif(qq,(abs(seqa(1,1,rows(qq))-maxindc(qq.>=ghat)).<n1));
  if _qnum > 0;
      nq=rows(qq);
      if nq>_qnum;
        qq=qq[ceil(seqa(1,1,_qnum)./_qnum*nq)];
      endif;
  endif;
retp(qq);
endp;


proc qfunc(x);
local q;
  q=x[.,(_dmin+1):(_dmax+1)];
retp(q);
endp;

proc (5) = tar2(dat);
local y,x,j,q,dhat,ghat,ghat1,ghat2,xx,e,qd,x1,f,gg,etar;
  y=dat[p+1:n];
  x=ones(t,1)~dat[p:n-1];
  j=2; do while j<=p;
    x=x~dat[p+1-j:n-j];
  j=j+1;endo;
  q = qfunc(x);
  {dhat,ghat}=tar(y,x,q,qsort(q));   @ fit TAR-1 model @
  qd=q[.,dhat];
  x1=x.*(qd.<=ghat); 
  xx=x1~x;
  e=y-xx*(y/xx);
  ghat2=tar_d(y,x,x1,qd,qsort_2(qd,ghat));  @ fit TAR-2 model @
  x1=x.*(qd.<=ghat2);
  ghat1=tar_d(y,x,x1,qd,qsort_2(qd,ghat2));  @ iterated estimate of gamma_1 @
  x1=x.*(qd.<=ghat1);
  ghat2=tar_d(y,x,x1,qd,qsort_2(qd,ghat1));  @ iterated estimate of gamma_2 @
  if ghat2<ghat1; gg=ghat1;ghat1=ghat2;ghat2=gg;endif;
  d1=(qd.<=ghat1);
  d2=(qd.<=ghat2).*(1-d1);
  xx=(x.*d1)~(x.*d2)~(x.*(1-d1-d2));
  etar=y-xx*(y/xx);
  f=t*((e'e)/(etar'etar)-1);
retp(f,dhat,ghat,ghat1,ghat2);
endp;


/*****************************************************/
@ Simulation Procedures @

proc tar_sim1(y0,b1,b2,delay,g,e);
local u,y,j,x,q;
  u=e[ceil(rndu(t,1)*t)];
  y=y0|zeros(t,1);
  j=1; do while j<=t;
    x=rev(y[j:p+j-1]);
    q=x[delay];
    x=1|x;
    y[p+j]=(b1'x)*(q<=g)+(b2'x)*(q>g)+u[j];
  j=j+1;endo;
retp(y);
endp;

proc tar_sim2(y0,b1,b2,delay,g,e1,e2);
local y,j,x,q,u1,u2,u,xx;
  y=y0|zeros(t,1);
  u=rndu(t,1);
  u1=e1[ceil(u*rows(e1))];
  u2=e2[ceil(u*rows(e2))];
  j=1; do while j<=t;
    x=rev(y[j:p+j-1]);
    xx=1|x;
    q=x[delay];
    y[p+j]=(b1'xx+u1[j])*(q<=g)+(b2'xx+u2[j])*(q>g);
  j=j+1;endo;
retp(y);
endp;

proc tar_sim3(y0,b1,b2,delay,g,het,z);
local y,u,j,x,q,xx,d1,d2,h;
  y=y0|zeros(t,1);
  u=z[ceil(rndu(t,1)*t)];
  j=1; do while j<=t;
    x=rev(y[j:p+j-1]);
    q=x[delay];
    xx=1|x;
    d1=(q<=g);
    d2=1-d1;
    h=((d1|d2|x).^2)'het;
    y[p+j]=(b1'xx)*d1+(b2'xx)*d2+sqrt(h.*(h.>0)).*u[j];
  j=j+1;endo;
retp(y);
endp;


/*****************************************************/  

proc kernel(x,b);
local h,g,kern,i,u;
  h=1.7*stdc(x)/(rows(x)^(.2));
  g=rows(b);
  kern=zeros(g,1);
  i=1; do while i<=g;
    u=abs(b[i]-x)/h;
    kern[i]=meanc((1-u.^2).*(u.<=1))*(.75)/h;
  i=i+1;endo;
retp(kern);
endp;

proc quant(x,q);
local s,qq;
  s=sortc(x,1);
  qq=s[round(rows(s)*q)];
retp(qq);
endp;

proc ginv(m);
local mi;
  trap 1;
  mi=invpd(m);
  trap 0;
  if scalerr(mi);
    mi=pinv(m);
  endif;
retp(mi);
endp;

proc dinv(y,x);
local d;
  trap 1;
    d=y/x;
  trap 0;
  if scalerr(d);
    d=ginv(moment(x,0))*(x'y);
  endif;
retp(d);
endp;

/*****************************************************/  