/* SUNSPOT2.PRG */

/* This replicates some of the empirical work reported in 
"Testing for Linearity" by Bruce E. Hansen.
For updates and contact information, see my webpage
www.ssc.wisc.edu/~bhansen
*/

/*  This program estimates a two-threshold TAR model, and tests the null of a 
one-regime AR against the alternative of a three-regime TAR */

load dat[]=sunspot.dat;
dat=(sqrt(1+dat)-1)*2;	@ load and define univariate series @
		@ load and define univariate series @
p=11;		@ order of autoregression @
_dmin=1;		@ minimal delay order @
_dmax=p;		@ maximal delay order (set to p or less @
_trim = .1;		@ minimal percentage of data per regime @
_qnum=100;	@ Number of thresholds to search. Set _qnum=0 to search over all values @
_boot = 2000;	@ bootstrap replications @
yname="Sunspots";	@ name of series @
output file = "sunspot2.out" reset; @ name of output file @

/*********************************************************************/

@ Define Data @
n=rows(dat);
y=dat[p+1:n];
t=n-p;
x=ones(t,1)~dat[p:n-1];
xname="Constant"|"Y(t-1)";
xname2="Constant"|"Y(t-1)^2";
j=2; do while j<=p;
    x=x~dat[p+1-j:n-j];
    pn = ftocv(j,1,0)*(j<10)+ftocv(j,2,0)*(j>=10);
    namej = "" $+ "Y(t-" $+ pn $+ ")";
    xname=xname|namej;
    xname2=xname2|(namej $+ "^2");
j=j+1;endo;
k=cols(x);

@ Linear Regression @
mi=invpd(moment(x,0));
beta=mi*(x'y);
e=y-x*beta;
ee=e'e;
xe=x.*e;
sig=ee/t;
se=sqrt(diag(mi*moment(xe,0)*mi));

@ Conditional Variance @
e2=e.^2;
x2 = x.^2;
m2=invpd(moment(x2,0));
hetbeta=m2*(x2'e2);
h=x2*hetbeta;
eh=e./sqrt(h.*(h.>0)+(h.<=0)).*(h.>0);
u=e2-h;
se2=sqrt(diag(m2*moment(x2.*u,0)*m2));
em=e2-meanc(e2);
fh=t*((em'em)/(u'u)-1);

@ Report Linear Estimation @
ff = "#*.*lG"~20~8;
print ("Dependent Variable:     " $+ yname);
"";
"Linear Autoregression";
"";
pr = printfm("Variable"~"Estimate"~"St Error",0~0~0,ff);"";
"       __________________________________________";
pr = printfm(xname~beta~se,0~1~1,ff);"";
"";
"Observations                      " t;
"Sum of Squared Errors             " ee;
"Residual Variance                 " sig;
"";"";
"Conditional Variance";"";
pr = printfm("Variable"~"Estimate"~"St Error",0~0~0,ff);"";
"       __________________________________________";
pr = printfm(xname2~hetbeta~se2,0~1~1,ff);"";
"";
"Heteroskedasticity F Test  " fh~cdfchic(fh,p);"";
"       __________________________________________";
"";"";
"";"";


@ TAR Estimation @

{f,dhat,ghat1,ghat2} = tar2(dat);

q = qfunc(x);
qd=q[.,dhat];
delay=_dmin-1+dhat;
d1=(qd.<=ghat1);
d2=(qd.<=ghat2).*(1-d1);
d3=1-d1-d2;
xx=(x.*d1)~(x.*d2)~(x.*d3);
mxx=invpd(moment(xx,0));
betatar=mxx*(xx'y);
etar=y-xx*betatar;
xxe=xx.*etar;
eetar=etar'etar;
sigtar=eetar/t;
n1=sumc(d1);n2=sumc(d2);n3=sumc(d3);
setar=sqrt(diag(mxx*moment(xxe,0)*mxx));
sig1=sumc((etar.*d1).^2)/n1;
sig2=sumc((etar.*d2).^2)/n2;
sig3=sumc((etar.*d3).^2)/n3;


@ Report TAR estimates @

"Threshold Autoregression";
"Sum of Squared Errors             " eetar;
"Residual Variance                 " sigtar;
"Delay Order                         " delay;
"First Threshold Estimate         " ghat1;
"Second Threshold Estimate       " ghat2;
"";"";
"Regime 1 (Threshold Variable less than or equal to First Threshold):";"";
pr = printfm("Variable"~"Estimate"~"St Error",0~0~0,ff);"";
"       __________________________________________";
pr = printfm(xname~betatar[1:k]~setar[1:k],0~1~1,ff);"";
"";
"Observations                      " n1;
"Percentage                         " (n1/t);
"Regime Variance                  " sig1;
"";"";
"Regime 2 (Threshold Variable Between Thresholds):";"";
pr = printfm("Variable"~"Estimate"~"St Error",0~0~0,ff);"";
"       __________________________________________";
pr = printfm(xname~betatar[k+1:2*k]~setar[k+1:2*k],0~1~1,ff);"";
"";
"Observations                      " n2;
"Percentage                         " (n2/t);
"Regime Variance                  " sig2;
"";"";
"Regime 3 (Threshold Variable Above Second Threshold):";"";
pr = printfm("Variable"~"Estimate"~"St Error",0~0~0,ff);"";
"       __________________________________________";
pr = printfm(xname~betatar[2*k+1:3*k]~setar[2*k+1:3*k],0~1~1,ff);"";
"";
"Observations                      " n3;
"Percentage                         " (n3/t);
"Regime Variance                  " sig3;
"       __________________________________________";
"";"";"";"";


@ Model-Based Bootstrap, Homoskedastic Errors @
  y0=dat[1:p];
  fb=zeros(_boot,1);
  i=1; do while i<=_boot;
   {fbi,dhatb,ghat1b,ghat2b} = tar2(ar_sim(y0,beta,e));
    fb[i]=fbi;
  i=i+1;endo;
  btpv=meanc(fb.>f);

@ Model-Based Bootstrap, Heteroskedastic Errors @
  y0=dat[1:p];
  fbh=zeros(_boot,1);
  i=1; do while i<=_boot;
   {fbi,dhatb,ghat1b,ghat2b} = tar2(het_sim(y0,beta,hetbeta,eh));
    fbh[i]=fbi;
  i=i+1;endo;
  btpvh=meanc(fbh.>f);

@ Output @
"Bootstrap Replications      :   " _boot;
names="Fstat"~"PV-BT"~"PV-BT-H";
statm=f~btpv~btpvh;
"";"";
ff = "#*.*lG"~12~7;
"TAR Tests, P-Values";"";
pr = printfm(names,0~0~0,ff);"";
"       __________________________________________";
ff = "#*.*lG"~12~4;
pr = printfm(statm,1~1~1,ff);"";
"";"";
output off;

@ Density Calculation @

ub=maxc(quant(fb,.99)|quant(fbh,.99));
k_x=seqa(0,ub/999,1000);
k_chi=(k_x.^((k)-1)).*exp(-k_x/2)./(gamma(k)*(2^k));
kern=k_chi~kernel(fb,k_x)~kernel(fbh,k_x);
xy(k_x,kern);
sun_k2=k_x~kern;
save sun_k2;


/*****************************************************/
@ TAR Procedures @

proc (2) = tar(y,x,q,qq);
local mi,e,qn,gn,s,m,qm,j,xd,xde,f,xxd,gg,mmi,dhat,sig,ghat;
  mi=ginv(moment(x,0));
  e=y-x*mi*(x'y);
  qn=cols(q);gn=rows(qq);
  s=zeros(gn,qn);
  m=1; do while m<=qn;
   qm=q[.,m]; 
    j=1;do while j<=gn;
      gg=qq[j,m];
      if scalmiss(gg); continue; endif;
      xd=x.*(qm.<=gg);
      xxd=moment(xd,0);
      mmi=xxd-xxd*mi*xxd;
      xde=xd'e;
      s[j,m]=xde'dinv(xde,mmi);
    j=j+1;endo;
  m=m+1;endo;
  dhat=maxindc(maxc(s));
  ghat=qq[maxindc(s[.,dhat]),dhat];
retp(dhat,ghat);
endp;

proc (2) = tar_d(y,x,x1,q,qq);
local mi,e,qn,gn,s,j,xd,xde,f,xxd,gg,mmi,ghat,xdx;
  mi=ginv(moment(x,0));
  e=y-x*mi*(x'y);
  gn=rows(qq);
  s=zeros(gn,1);
    j=1;do while j<=gn;
      gg=qq[j];
      if scalmiss(gg); continue; endif;
      xd=x1~(x.*(q.<=gg));
      xxd=moment(xd,0);
      xdx=x'xd;
      mmi=xxd-xdx'mi*xdx;
      xde=xd'e;
      s[j]=xde'dinv(xde,mmi);
    j=j+1;endo;
  f=maxc(s);
  ghat=qq[maxindc(s)];
retp(f,ghat);
endp;


proc qsort(q);
local j,n,n1,qq,k,qj,nj;
  n=rows(q);k=cols(q);
  n1=round(_trim*n);
  if _qnum==0;
    qq=zeros(n-n1-n1+1,k);
    qq=miss(qq,0);
    j=1; do while j<=k;
      qj=unique(q[.,j],1);
      qj=qj[n1:(rows(qj)-n1+1)];
      qq[1:rows(qj),j]=qj;
    j=j+1;endo;
  else;
    qq=zeros(_qnum,k);
    qq=miss(qq,0);
    j=1; do while j<=k;
      qj=unique(q[.,j],1);
      qj=qj[n1:(rows(qj)-n1+1)];
      nj=rows(qj);
      if nj<=_qnum;
        qq[1:nj,j]=qj;
      else;
        qq[.,j]=qj[ceil(seqa(1,1,_qnum)./_qnum*nj)];
      endif;
    j=j+1;endo;
  endif;
retp(qq);
endp;


proc qsort_2(q,ghat);
local n,n1,qq,nq;
  n=rows(q);
  n1=round(_trim*n);
  qq=unique(q,1);
  qq=qq[n1:(rows(qq)-n1+1)];
  qq=delif(qq,(abs(seqa(1,1,rows(qq))-maxindc(qq.>=ghat)).<n1));
  if _qnum > 0;
      nq=rows(qq);
      if nq>_qnum;
        qq=qq[ceil(seqa(1,1,_qnum)./_qnum*nq)];
      endif;
  endif;
retp(qq);
endp;


proc qfunc(x);
local q;
  q=x[.,(_dmin+1):(_dmax+1)];
retp(q);
endp;

proc (4) = tar2(dat);
local y,x,j,q,dhat,ghat1,ghat2,e,qd,x1,f,gg;
  y=dat[p+1:n];
  x=ones(t,1)~dat[p:n-1];
  j=2; do while j<=p;
    x=x~dat[p+1-j:n-j];
  j=j+1;endo;
  e=y-x*(y/x);
  q = qfunc(x);
  {dhat,ghat1}=tar(y,x,q,qsort(q));   @ fit TAR-1 model @
  qd=q[.,dhat];
  x1=x.*(qd.<=ghat1);
  {f,ghat2}=tar_d(y,x,x1,qd,qsort_2(qd,ghat1));  @ fit TAR-2 model @
  x1=x.*(qd.<=ghat2);
  {f,ghat1}=tar_d(y,x,x1,qd,qsort_2(qd,ghat2));  @ iterated estimate of gamma_1 @
  x1=x.*(qd.<=ghat1);
  {f,ghat2}=tar_d(y,x,x1,qd,qsort_2(qd,ghat1));  @ iterated estimate of gamma_2 @
   f=t/(((e'e)/f)-1);
   if ghat2<ghat1; gg=ghat1;ghat1=ghat2;ghat2=gg;endif;
retp(f,dhat,ghat1,ghat2);
endp;


/*****************************************************/
@ Simulation Procedures @

proc ar_sim(y0,beta,e);
local t,u,yb;
  t=rows(e);
  u=e[ceil(rndu(t,1)*t)]+beta[1];
  yb=recserar(y0|u,y0,beta[2:k]);
retp(yb);
endp;


proc het_sim(y0,beta,het,e);
local t,u,j,y,x,h;
  t=rows(e);
  u=e[ceil(rndu(t,1)*t)];
  y=y0|zeros(t,1);
  j=1; do while j<=t;
    x=1|rev(y[j:p+j-1]);
    h=(x.^2)'het;
    y[p+j]=beta'x+sqrt(h.*(h.>0)).*u[j];
  j=j+1;endo;
retp(y);
endp;

/*****************************************************/  

proc kernel(x,b);
local h,g,kern,i,u;
  h=1.7*stdc(x)/(rows(x)^(.2));
  g=rows(b);
  kern=zeros(g,1);
  i=1; do while i<=g;
    u=abs(b[i]-x)/h;
    kern[i]=meanc((1-u.^2).*(u.<=1))*(.75)/h;
  i=i+1;endo;
retp(kern);
endp;

proc quant(x,q);
local s,qq;
  s=sortc(x,1);
  qq=s[round(rows(s)*q)];
retp(qq);
endp;

proc ginv(m);
local mi;
  trap 1;
  mi=invpd(m);
  trap 0;
  if scalerr(mi);
    mi=pinv(m);
  endif;
retp(mi);
endp;

proc dinv(y,x);
local d;
  trap 1;
    d=y/x;
  trap 0;
  if scalerr(d);
    d=ginv(moment(x,0))*(x'y);
  endif;
retp(d);
endp;

/*****************************************************/  